#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

"""
Definitions for DRMAA2 exception classes.
"""

from .drmaa2_constants import StatusCode


class Drmaa2Exception(Exception):
    """ 
    Base class for all DRMAA2 exceptions. 

    Usage examples:
    
    >>> raise Drmaa2Exception(error_message, error_code)
    """

    def __init__(self, error='', error_code=StatusCode.UNSET_ERROR):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str

        :param error_code: Error code.
        :type error_code: int
        """
        Exception.__init__(self, error)
        self.error_code = error_code


class DeniedByDrms(Drmaa2Exception):
    """ DRM denied the operation. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.DENIED_BY_DRMS)


class DrmCommunicationError(Drmaa2Exception):
    """ DRM could not be contacted. The problem may or may not be transient. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.DRM_COMMUNICATION)


class TryLaterError(Drmaa2Exception):
    """ An error due to a transient problem that can be retried. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.TRY_LATER)


class SessionManagementError(Drmaa2Exception):
    """ Session management error. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.SESSION_MANAGEMENT)


class TimeoutError(Drmaa2Exception):
    """ Operation timed out. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.TIMEOUT)


class InternalError(Drmaa2Exception):
    """ DRMAA2 Internal error. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.INTERNAL)


class InvalidArgument(Drmaa2Exception):
    """ Invalid argument error. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.INVALID_ARGUMENT)


class InvalidSession(Drmaa2Exception):
    """ Invalid (e.g., previously closed) session used for the operation. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.INVALID_SESSION)


class InvalidState(Drmaa2Exception):
    """ Operation is not allowed given the current state of the job. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.INVALID_STATE)


class ResourceNotAvailable(Drmaa2Exception):
    """ Requested resource is not available. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.OUT_OF_RESOURCE)


class UnsupportedAttribute(Drmaa2Exception):
    """ The given attribute is not supported by the DRMAA2 implementation. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.UNSUPPORTED_ATTRIBUTE)


class UnsupportedOperation(Drmaa2Exception):
    """ The given operation is not supported by the DRMAA2 implementation. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.UNSUPPORTED_OPERATION)


class ImplementationSpecificError(Drmaa2Exception):
    """ Implementation-specific error condition. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.IMPLEMENTATION_SPECIFIC)


class AuthorizationError(Drmaa2Exception):
    """ Authorization error. """

    def __init__(self, error=''):
        """ 
        Class constructor. 

        :param error: Error message.
        :type error: str
        """
        Drmaa2Exception.__init__(self, error=error, error_code=StatusCode.AUTHORIZATION)



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

import platform
from .drmaa2_constants import PY_STRING_TYPE


class ByteString:
    python_version = platform.python_version_tuple()[0]

    def __init__(self, s=''):
        if isinstance(s, ByteString):
            self.s = s.s
        self.s = s

    def __str__(self):
        return str(self.s)

    def encode(self):
        if self.s is None or self.python_version == '2':
            return self.s
        else:
            try:
                return self.s.encode()
            except AttributeError:
                return self.s

    def decode(self):
        if self.s is None or self.python_version == '2':
            return self.s
        else:
            try:
                return self.s.decode()
            except AttributeError:
                return self.s



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

"""
Definitions for various DRMAA2 constants, including enums.
"""

from datetime import datetime
from enum import Enum
from enum import IntEnum

UNSET_BOOL = 0
UNSET_ENUM = -1
UNSET_NUM = -1
UNSET_PRIORITY = -99999
UNSET_TIME = -3

POSIX_EPOCH = datetime.utcfromtimestamp(0)
PY_DICT_TYPE = type(dict())
PY_STRING_TYPE = type('string')
PY_BYTES_TYPE = type(b'string')


class Bool(IntEnum):
    """ DRMAA2 Bool enum. """
    FALSE = 0
    TRUE = 1


class Time(IntEnum):
    """ DRMAA2 Time enum. """
    ZERO_TIME = 0
    INFINITE_TIME = -1
    NOW = -2
    UNSET_TIME = -3


class Capability(IntEnum):
    """ DRMAA2 Capabilities enum. """
    UNSET_CAPABILITY = -1
    ADVANCE_RESERVATION = 0
    RESERVE_SLOTS = 1
    CALLBACK = 2
    BULK_JOBS_MAXPARALLEL = 3
    JT_EMAIL = 4
    JT_STAGING = 5
    JT_DEADLINE = 6
    JT_MAXSLOTS = 7
    JT_ACCOUNTINGID = 8
    RT_STARTNOW = 9
    RT_DURATION = 10
    RT_MACHINEOS = 11
    RT_MACHINEARCH = 12


class Cpu(IntEnum):
    """ DRMAA2 CPU enum. """
    UNSET_CPU = -1
    OTHER_CPU = 0
    ALPHA = 1
    ARM = 2
    ARM64 = 3
    CELL = 4
    PARISC = 5
    PARISC64 = 6
    X86 = 7
    X64 = 8
    IA64 = 9
    MIPS = 10
    MIPS64 = 11
    PPC = 12
    PPC64 = 13
    SPARC = 14
    SPARC64 = 15
    PPC64LE = 16


class Event(IntEnum):
    """ DRMAA2 event enum. """
    UNSET_EVENT = -1
    NEW_STATE = 0
    MIGRATED = 1
    ATTRIBUTE_CHANGE = 2


class JobState(IntEnum):
    """ DRMAA2 job state enum. """
    UNSET_JSTATE = -1
    UNDETERMINED = 0
    QUEUED = 1
    QUEUED_HELD = 2
    RUNNING = 3
    SUSPENDED = 4
    REQUEUED = 5
    REQUEUED_HELD = 6
    DONE = 7
    FAILED = 8
    RUNNING_HELD = 9


class Os(IntEnum):
    """ DRMAA2 OS enum. """
    UNSET_OS = -1
    OTHER_OS = 0
    AIX = 1
    BSD = 2
    LINUX = 3
    HPUX = 4
    IRIX = 5
    MACOS = 6
    SUNOS = 7
    TRU64 = 8
    UNIXWARE = 9
    WIN = 10
    WINNT = 11


class ResourceLimit(Enum):
    """ DRMAA2 resource limits enum. """
    CORE_FILE_SIZE = "CORE_FILE_SIZE"
    CPU_TIME = "CPU_TIME"
    DATA_SIZE = "DATA_SIZE"
    FILE_SIZE = "FILE_SIZE"
    OPEN_FILES = "OPEN_FILES"
    STACK_SIZE = "STACK_SIZE"
    VIRTUAL_MEMORY = "VIRTUAL_MEMORY"
    WALLCLOCK_TIME = "WALLCLOCK_TIME"


class ListType(IntEnum):
    """ DRMAA2 list type enum. """
    UNSET_LISTTYPE = -1
    STRINGLIST = 0
    JOBLIST = 1
    QUEUEINFOLIST = 2
    MACHINEINFOLIST = 3
    SLOTINFOLIST = 4
    RESERVATIONLIST = 5


class StatusCode(IntEnum):
    """ DRMAA2 status code enum. """
    UNSET_ERROR = -1
    SUCCESS = 0
    DENIED_BY_DRMS = 1
    DRM_COMMUNICATION = 2
    TRY_LATER = 3
    SESSION_MANAGEMENT = 4
    TIMEOUT = 5
    INTERNAL = 6
    INVALID_ARGUMENT = 7
    INVALID_SESSION = 8
    INVALID_STATE = 9
    OUT_OF_RESOURCE = 10
    UNSUPPORTED_ATTRIBUTE = 11
    UNSUPPORTED_OPERATION = 12
    IMPLEMENTATION_SPECIFIC = 13
    AUTHORIZATION = 14
    LASTERROR = 15



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from ctypes import POINTER
from ctypes import pointer

from .drmaa2_constants import UNSET_PRIORITY
from .drmaa2_constants import Os
from .drmaa2_constants import Cpu
from .drmaa2_ctypes import drmaa2_jtemplate
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_object import Drmaa2Object
from .drmaa2_exceptions import InvalidArgument


class JobTemplate(Drmaa2Object):
    """ High-level DRMAA2 job template class. """

    remote_command = Drmaa2Object.StringDescriptor('remoteCommand')
    """ Job executable (str). """
    args = Drmaa2Object.StringListDescriptor('args')
    """ Arguments for the job executable ([str]). """
    submit_as_hold = Drmaa2Object.BoolDescriptor('submitAsHold')
    """ Submit as hold flag (bool). """
    rerunnable = Drmaa2Object.BoolDescriptor('rerunnable')
    """ Rerrunable flag (bool). """
    job_environment = Drmaa2Object.DictDescriptor('jobEnvironment')
    """ Job environment variables (dictionary). """
    working_directory = Drmaa2Object.StringDescriptor('workingDirectory')
    """ Working directory (str). """
    job_category = Drmaa2Object.StringDescriptor('jobCategory')
    """ Job category (str). """
    email = Drmaa2Object.StringListDescriptor('email')
    """ List of emails for notifications ([str]). """
    email_on_started = Drmaa2Object.BoolDescriptor('emailOnStarted')
    """ Email on started flag (bool). """
    email_on_terminated = Drmaa2Object.BoolDescriptor('emailOnTerminated')
    """ Email on terminated flag (bool). """
    job_name = Drmaa2Object.StringDescriptor('jobName')
    """ Job name (str). """
    input_path = Drmaa2Object.StringDescriptor('inputPath')
    """ Input file path (str). """
    output_path = Drmaa2Object.StringDescriptor('outputPath')
    """ Output file path (str). """
    error_path = Drmaa2Object.StringDescriptor('errorPath')
    """ Error file path (str). """
    join_files = Drmaa2Object.BoolDescriptor('joinFiles')
    """ Join files flag (bool). """
    reservation_id = Drmaa2Object.StringDescriptor('reservationId')
    """ Reservation id (str). """
    queue_name = Drmaa2Object.StringDescriptor('queueName')
    """ Requested queue (str). """
    min_slots = Drmaa2Object.LongLongDescriptor('minSlots')
    """ Minimum number of slots required by the job (long). """
    max_slots = Drmaa2Object.LongLongDescriptor('maxSlots')
    """ Maximum number of slots required by the job (long). """
    priority = Drmaa2Object.LongLongDescriptor('priority', unset_value=UNSET_PRIORITY)
    """ Job priority (long). """
    candidate_machines = Drmaa2Object.StringListDescriptor('candidateMachines')
    """ List of machines that are candidate for running the job ([str]). """
    min_phys_memory = Drmaa2Object.LongLongDescriptor('minPhysMemory')
    """ Minimum physical memory needed for the job (long). """
    machine_os = Drmaa2Object.EnumDescriptor('machineOS', Os)
    """ OS required for the job (Os). """
    machine_arch = Drmaa2Object.EnumDescriptor('machineArch', Cpu)
    """ Machine arch required for the job (Cpu). """
    start_time = Drmaa2Object.TimeDescriptor('startTime')
    """ Requested start time (datetime). """
    deadline_time = Drmaa2Object.TimeDescriptor('deadlineTime')
    """ Deadline time (datetime). """
    stage_in_files = Drmaa2Object.DictDescriptor('stageInFiles')
    """ Stage in files ({str:str}). """
    stage_out_files = Drmaa2Object.DictDescriptor('stageOutFiles')
    """ Stage out files ({str:str}). """
    resource_limits = Drmaa2Object.DictDescriptor('resourceLimits')
    """ Resource limits ({str:str}). """
    accounting_id = Drmaa2Object.StringDescriptor('accountingId')
    """ Accounting id (str). """
    implementation_specific = Drmaa2Object.ImplSpecDescriptor('implementationSpecific')
    """ Implementation specific dictionary ({str:str}). """

    def __init__(self, template={}):
        """ 
        Constructor. 

        :param template: Input structure representing the object; this structure is typically a dictionary, but it can also be a low-level drmaa2_jtemplate struct.
        :type template: dict 

        :raises InvalidArgument: in case of an invalid input argument.
        :raises Drmaa2Exception: for all other errors.

        >>> jt = JobTemplate({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> print(jt.remote_command)
        /bin/sleep
        >>> jt.rerunnable = True
        >>> print(jt.get_implementation_specific_keys())
        ['uge_jt_pe', 'uge_jt_native']
        """
        Drmaa2Object.__init__(self)
        if isinstance(template, dict):
            self._struct = self.get_drmaa2_library().drmaa2_jtemplate_create()
            # self._struct = POINTER(drmaa2_jtemplate)
            # self._struct.contents = drmaa2_jtemplate()
            self.init_impl_spec_key_values()
            self.from_dict(template)
        elif isinstance(template, POINTER(drmaa2_jtemplate)):
            self._struct = template
        else:
            raise InvalidArgument('Invalid argument: %s' % str(template))

    def __del__(self):
        setattr(self._struct.contents, 'remoteCommand', drmaa2_string())
        setattr(self._struct.contents, 'workingDirectory', drmaa2_string())
        setattr(self._struct.contents, 'jobCategory', drmaa2_string())
        setattr(self._struct.contents, 'jobName', drmaa2_string())
        setattr(self._struct.contents, 'inputPath', drmaa2_string())
        setattr(self._struct.contents, 'outputPath', drmaa2_string())
        setattr(self._struct.contents, 'errorPath', drmaa2_string())
        setattr(self._struct.contents, 'reservationId', drmaa2_string())
        setattr(self._struct.contents, 'queueName', drmaa2_string())
        setattr(self._struct.contents, 'job_environment', dict())
        setattr(self._struct.contents, 'stage_in_files', dict())
        setattr(self._struct.contents, 'stage_out_files', dict())
        setattr(self._struct.contents, 'resource_limits', dict())
        setattr(self._struct.contents, 'accountingId', drmaa2_string())
        self.get_drmaa2_library().drmaa2_jtemplate_free(pointer(self._struct))

    @classmethod
    def get_implementation_specific_keys(cls):
        """
        Retrieve list of implementation-specific keys.

        :returns: String list of implementation-specific keys.
        """
        if cls.implementation_specific_keys is None:
            cls.implementation_specific_keys = cls.to_py_string_list(
                cls.get_drmaa2_library().drmaa2_jtemplate_impl_spec())
        return cls.implementation_specific_keys



#!/usr/bin/env python 
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

import getpass
from uuid import uuid4
from ctypes import pointer
from ctypes import c_void_p

from .byte_string import ByteString
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_object import Drmaa2Object

from .sudo import Sudo
from .reservation_template import ReservationTemplate
from .reservation import Reservation
from .log_manager import LogManager
from .exception_mapper import ExceptionMapper
from .drmaa2_exceptions import Drmaa2Exception


class ReservationSession(Drmaa2Object):
    """ High-level DRMAA2 reservation session class. """

    name = Drmaa2Object.StringDescriptor('name')
    """ Reservation session name (str). """
    contact = Drmaa2Object.StringDescriptor('contact')
    """ Reservation session contact (str). """

    logger = LogManager.get_instance().get_logger('ReservationSession')
    exception_mapper = ExceptionMapper()

    def __init__(self, name=None, contact=None, destroy_on_exit=True, check_for_existing_session=True, session_dict={}, auth=None):
        """ 
        Constructor. If the session with a given name does not already exist, 
        it will create a new one; otherwise, it will open the existing session.

        :param name: Session name; it may be also be specified as part of the reservation session object dictionary. If not provided, a random name will be used.
        :type name: str

        :param contact: Session contact; it may be also be specified as part of the reservation session object dictionary.
        :type contact: str

        :param destroy_on_exit: Destroy session when the reservation session object goes out of scope.
        :type destroy_on_exit: bool

        :param check_for_existing_session: If false, the constructor will not check for existing session with the given name.
        :type check_for_existing_session: bool

        :param session_dict: Reservation session dictionary; if specified, it should contain 'name' and (optionally) 'contact' keys.
        :type session_dict: dict

        :param auth: Optional sudo object for creating the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> r_session = ReservationSession()
        >>> print(r_session.name)
        def678952a654dfda75ffabae6890be
        """
        Drmaa2Object.__init__(self)
        self._destroy_on_exit = destroy_on_exit
        name = name or session_dict.get('name')
        if not name:
            name = uuid4().hex
            check_for_existing_session = False
        contact = contact or session_dict.get('contact') or drmaa2_string()
        create_new_session = True
        if check_for_existing_session:
             existing_session_names = self.list_session_names()
             session_names_to_check = [name]
             if contact:
                 session_names_to_check.append('%s@%s' % (contact,name))
             else:
                 session_names_to_check.append('%s@%s' % (getpass.getuser(),name))
             for n in session_names_to_check:
                 if n in existing_session_names:
                     name = n
                     self.logger.debug('Discovered existing job session with name {}'.format(n))
                     create_new_session = False
                     break

        self._name_bs = ByteString()
        self._auth = auth
        if create_new_session:
            self._struct = self.__create(name, contact, auth)
        else:
            self._struct = self.__open(name)
        self._struct_p = pointer(self._struct)
        self._name_bs = ByteString(name)
        self._read_only = True

    @classmethod
    def list_session_names(cls):
        """ 
        List existing sessions. 

        :returns: List of existing reservation session names.

        >>> print(ReservationSession.list_session_names())
        ['auser@rs-01']

        """
        return Drmaa2Object.to_py_string_list(cls.get_drmaa2_library().drmaa2_get_rsession_names())

    def __create(self, name, contact, auth):
        self.logger.debug('Creating reservation session with name {} (contact: {})'.format(name, contact))
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            struct = self.get_drmaa2_library().drmaa2_create_rsession_as(auth._struct, ByteString(name).encode(), ByteString(contact).encode())
        else:
            struct = self.get_drmaa2_library().drmaa2_create_rsession(ByteString(name).encode(), ByteString(contact).encode())
        if not struct:
            self.exception_mapper.check_last_error_code()
        return struct

    def __open(self, name):
        self.logger.debug('Opening reservation session {}'.format(name))
        struct = self.get_drmaa2_library().drmaa2_open_rsession(ByteString(name).encode())
        if not struct:
            self.exception_mapper.check_last_error_code()
        return struct

    def open(self):
        """ 
        Open session. This method is called automatically in the constructor.

        >>> ...
        >>> r_session.close()
        >>> ...
        >>> r_session.open()
        """
        if self._struct:
            self.logger.debug('Reservation session {} is already open'.format(self._name_bs.decode()))
        else:
            self._struct = self.__open(self._name_bs)
            self._struct_p = pointer(self._struct)

    def close(self):
        """ 
        Close session. This method is called automatically when the session
        object goes out of scope.

        >>> ...
        >>> r_session.close()
        """
        if self._struct:
            self.logger.debug('Closing reservation session {}'.format(self._name_bs.decode()))
            drmaa2_lib = self.get_drmaa2_library()
            self.exception_mapper.check_status_code(drmaa2_lib.drmaa2_close_rsession(self._struct))
            drmaa2_lib.drmaa2_rsession_free(self._struct_p)
            self._struct = None

    def destroy(self, auth=None):
        """ 
        Destroy session. This method is called automatically when the 
        session object goes out of scope, unless the destroy_on_exit flag is set to false.

        :param auth: Optional sudo object for destroying the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> ...
        >>> r_session.destroy()
        """
        self.logger.debug('Destroying reservation session {}'.format(self._name_bs.decode()))
        auth = auth or self._auth
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            self.exception_mapper.check_status_code(self.get_drmaa2_library().drmaa2_destroy_rsession_as(auth._struct, self._name_bs.encode()))
        else:
            self.exception_mapper.check_status_code(self.get_drmaa2_library().drmaa2_destroy_rsession(self._name_bs.encode()))

    @classmethod
    def destroy_by_name(cls, name, auth=None):
        """ 
        Destroy session by name. 

        :param name: Session name.
        :type name: str

        :param auth: Optional sudo object for destroying the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> ReservationSession.destroy_by_name('rs-01')
        """
        cls.logger.debug('Destroying reservation session {}'.format(name))
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            cls.exception_mapper.check_status_code(cls.get_drmaa2_library().drmaa2_destroy_rsession_as(auth._struct, ByteString(name).encode()))
        else:
            cls.exception_mapper.check_status_code(cls.get_drmaa2_library().drmaa2_destroy_rsession(ByteString(name).encode()))

    def __del__(self):
        """ Destructor. """
        self.close()
        if self._destroy_on_exit:
            try:
                self.destroy()
            except Drmaa2Exception as ex:
                self.logger.warn('Could not destroy reservation session: {}'.format(str(ex)))
        else:
            self.logger.debug('Will not destroy reservation session {}'.format(self._name_bs.decode()))
        if self._struct:
            self.get_drmaa2_library().drmaa2_rsession_free(self._struct_p)

    def request_reservation(self, template, auth=None):
        """ 
        Request reservation. 

        :param template: Reservation template; it can be specified either as a dictionary, or directly as a ReservationTemplate object.
        :type template: ReservationTemplate or dict

        :param auth: Optional sudo object for requesting the reservation; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        :returns: Reservation object.

        >>> r = r_session.request_reservation({'reservation_name' : 'res-01', 'duration' : 100})
        >>> print(r.id)
        623
        """
        self.logger.debug('Requesting a reservation using template: {}'.format(template))
        drmaa2_lib = self.get_drmaa2_library()
        template = ReservationTemplate.create_from_dict(template)
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ctypes_reservation = drmaa2_lib.drmaa2_rsession_request_reservation_as(auth._struct, self._struct, template._struct)
        else:
            ctypes_reservation = drmaa2_lib.drmaa2_rsession_request_reservation(self._struct, template._struct)
        if not ctypes_reservation:
            self.exception_mapper.check_last_error_code()
        py_reservation = Reservation(ctypes_reservation)
        self.logger.debug('Got reservation {}'.format(py_reservation))
        drmaa2_lib.drmaa2_r_free(pointer(ctypes_reservation))
        return py_reservation

    def get_reservation(self, id):
        """ 
        Get reservation. 

        :param id: Reservation id.
        :type id: str

        :returns: Reservation object.

        >>> r = r_session.get_reservation('623')
        >>> print(r.id)
        623
        """
        self.logger.debug('Requesting reservation id: {}'.format(id))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_reservation = drmaa2_lib.drmaa2_rsession_get_reservation(self._struct, ByteString(id).encode())
        if not ctypes_reservation:
            self.exception_mapper.check_last_error_code()
        py_reservation = Reservation(ctypes_reservation)
        self.logger.debug('Got reservation {}'.format(py_reservation))
        drmaa2_lib.drmaa2_r_free(pointer(ctypes_reservation))
        return py_reservation

    def get_reservations(self):
        """ 
        Get all reservations. 

        :returns: List of Reservation objects.

        >>> r_list = r_session.get_reservations()
        """
        self.logger.debug('Requesting all reservations')
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_reservation_list = drmaa2_lib.drmaa2_rsession_get_reservations(self._struct)
        if not ctypes_reservation_list:
            self.exception_mapper.check_last_error_code()
        py_reservation_list = Reservation.to_py_reservation_list(ctypes_reservation_list)
        self.logger.debug('Retrieved {} reservations'.format(len(py_reservation_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_reservation_list)))
        return py_reservation_list




#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from .version import Version
from .drmaa2_object_descriptors import Drmaa2Descriptor

from .log_manager import LogManager


class Drmaa2VersionDescriptor(Drmaa2Descriptor):
    """ A descriptor for drmaa2_version fields. """

    logger = LogManager.get_instance().get_logger('Drmaa2VersionDescriptor')

    def __init__(self, name):
        Drmaa2Descriptor.__init__(self, name)

    def __get__(self, obj, type=None):
        if not obj:
            return None
        if obj._struct is not None:
            ctypes_version = getattr(obj._struct.contents, self.name)
            py_version = Version(ctypes_version)
        return py_version

    def __set__(self, obj, ctypes_version):
        if not obj:
            return
            # setattr(obj._struct.contents, self.name, value)



#!/usr/bin/env python 
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

import getpass
from uuid import uuid4
from ctypes import pointer
from ctypes import c_void_p
from .drmaa2_constants import Time
from .drmaa2_constants import PY_DICT_TYPE

from .byte_string import ByteString
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_object import Drmaa2Object

from .sudo import Sudo
from .job_template import JobTemplate
from .job import Job
from .job_info import JobInfo
from .job_array import JobArray
from .log_manager import LogManager
from .exception_mapper import ExceptionMapper
from .drmaa2_exceptions import Drmaa2Exception
from .drmaa2_exceptions import InvalidArgument


class JobSession(Drmaa2Object):
    """ High-level DRMAA2 job session class. """

    name = Drmaa2Object.StringDescriptor('name')
    """ Job session name (str). """
    contact = Drmaa2Object.StringDescriptor('contact')
    """ Job session contact (str). """

    logger = LogManager.get_instance().get_logger('JobSession')
    exception_mapper = ExceptionMapper()

    def __init__(self, name=None, contact=None, destroy_on_exit=True, check_for_existing_session=True, session_dict={},
                 auth=None):
        """ 
        Constructor. If the session with a given name does not already exist, 
        it will create a new one; otherwise, it will open the existing session.

        :param name: Session name; it may be also be specified as part of the job session object dictionary. If not provided, a random name will be used.
        :type name: str

        :param contact: Session contact; it may be also be specified as part of the job session object dictionary.
        :type contact: str

        :param destroy_on_exit: Destroy session when the job session object goes out of scope.
        :type destroy_on_exit: bool

        :param check_for_existing_session: If false, the constructor will not check for existing session with the given name.
        :type check_for_existing_session: bool

        :param session_dict: Job session dictionary; if specified, it should contain 'name' and (optionally) 'contact' keys.
        :type session_dict: dict

        :param auth: Optional sudo object for creating the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j_session = JobSession()
        >>> print(j_session.name)
        bab80df52a654dfda7552cbae8dacca
        """
        Drmaa2Object.__init__(self)
        self._destroy_on_exit = destroy_on_exit
        name = name or session_dict.get('name')
        if not name:
            name = uuid4().hex
            check_for_existing_session = False

        contact = contact or session_dict.get('contact') or drmaa2_string()
        create_new_session = True
        if check_for_existing_session:
            existing_session_names = self.list_session_names()
            session_names_to_check = [name]
            if contact:
                session_names_to_check.append('%s@%s' % (contact, name))
            else:
                session_names_to_check.append('%s@%s' % (getpass.getuser(), name))
            for n in session_names_to_check:
                if n in existing_session_names:
                    name = n
                    self.logger.debug('Discovered existing job session with name {}'.format(n))
                    create_new_session = False
                    break

        self._name_bs = ByteString()
        self._auth = auth
        if create_new_session:
            self._struct = self.__create(name, contact, auth)
        else:
            self._struct = self.__open(name)
        self._struct_p = pointer(self._struct)
        self._name_bs = ByteString(name)
        self._read_only = True

    @classmethod
    def list_session_names(cls):
        """ 
        List existing sessions. 

        :returns: List of existing job session names.

        >>> print(JobSession.list_session_names())
        ['auser@js-01']
        """
        return Drmaa2Object.to_py_string_list(cls.get_drmaa2_library().drmaa2_get_jsession_names())

    def __create(self, name, contact, auth):
        self.logger.debug('Creating job session with name {} (contact: {})'.format(name, contact))
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            struct = self.get_drmaa2_library().drmaa2_create_jsession_as(auth._struct, ByteString(name).encode(),
                                                                         ByteString(contact).encode())
        else:
            if '@' in name:
                raise InvalidArgument('session name with @ is not allowed')
            struct = self.get_drmaa2_library().drmaa2_create_jsession(ByteString(name).encode(),
                                                                      ByteString(contact).encode())
        if not struct:
            self.exception_mapper.check_last_error_code()
        return struct

    def __open(self, name):
        self.logger.debug('Opening job session {}'.format(name))
        struct = self.get_drmaa2_library().drmaa2_open_jsession(ByteString(name).encode())
        if not struct:
            self.exception_mapper.check_last_error_code()
        return struct

    def open(self):
        """ 
        Open session. This method is called automatically in the constructor.

        >>> ...
        >>> j_session.close()
        >>> ...
        >>> j_session.open()
        """
        if self._struct:
            self.logger.debug('Job session {} is already open'.format(self._name_bs.decode()))
        else:
            self._struct = self.__open(self._name_bs)
            self._struct_p = pointer(self._struct)

    def close(self):
        """ 
        Close session. This method is called automatically when the session
        object goes out of scope.

        >>> ...
        >>> j_session.close()
        """
        if self._struct:
            self.logger.debug('Closing job session {}'.format(self._name_bs.decode()))
            drmaa2_lib = self.get_drmaa2_library()
            self.exception_mapper.check_status_code(drmaa2_lib.drmaa2_close_jsession(self._struct))
            drmaa2_lib.drmaa2_jsession_free(self._struct_p)
            self._struct = None

    def destroy(self, auth=None):
        """ 
        Destroy session. This method is called automatically when the 
        session object goes out of scope, unless the destroy_on_exit flag is set to false.

        :param auth: Optional sudo object for destroying the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> ...
        >>> j_session.destroy()
        """
        self.logger.debug('Destroying job session {}'.format(self._name_bs.decode()))
        auth = auth or self._auth
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            if self._name_bs.decode() != ByteString('').decode():
                self.exception_mapper.check_status_code(
                    self.get_drmaa2_library().drmaa2_destroy_jsession_as(auth._struct, self._name_bs.encode()))
        else:
            if self._name_bs.decode() != ByteString('').decode():
                self.exception_mapper.check_status_code(
                    self.get_drmaa2_library().drmaa2_destroy_jsession(self._name_bs.encode()))

    @classmethod
    def destroy_by_name(cls, name, auth=None):
        """ 
        Destroy session by name. 

        :param name: Session name.
        :type name: str

        :param auth: Optional sudo object for destroying the session; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> JobSession.destroy_by_name('js-01')
        """
        cls.logger.debug('Destroying job session {}'.format(name))
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            cls.exception_mapper.check_status_code(
                cls.get_drmaa2_library().drmaa2_destroy_jsession_as(auth._struct, ByteString(name).encode()))
        else:
            cls.exception_mapper.check_status_code(
                cls.get_drmaa2_library().drmaa2_destroy_jsession(ByteString(name).encode()))

    def __del__(self):
        """ Destructor. """
        self.close()
        if self._destroy_on_exit:
            try:
                self.destroy()
            except Drmaa2Exception as ex:
                self.logger.warn('Could not destroy job session: {}'.format(str(ex)))
        else:
            self.logger.debug('Will not destroy job session {}'.format(self._name_bs.decode()))
        if self._struct:
            self.get_drmaa2_library().drmaa2_jsession_free(self._struct_p)

    def run_job(self, template, auth=None):
        """ 
        Run a job using template. 

        :param template: Job template; it can be specified either as a dictionary, or directly as a JobTemplate object.
        :type template: JobTemplate or dict

        :param auth: Optional sudo object for running the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        :returns: Job object.

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['10']})
        >>> print(j)
        {'id': '521', 'session_name': 'js-01'}
        >>> type(j)
        <class 'drmaa2.job.Job'>
        >>> print(j.id)
        521
        """
        self.logger.debug('Running a job using template: {}'.format(template))
        drmaa2_lib = self.get_drmaa2_library()
        template = JobTemplate.create_from_dict(template)
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ctypes_job = drmaa2_lib.drmaa2_jsession_run_job_as(auth._struct, self._struct, template._struct)
        else:
            ctypes_job = drmaa2_lib.drmaa2_jsession_run_job(self._struct, template._struct)
        if not ctypes_job:
            self.exception_mapper.check_last_error_code()
        py_job = Job(ctypes_job)
        self.logger.debug('Got job {}'.format(py_job))
        drmaa2_lib.drmaa2_j_free(pointer(ctypes_job))
        return py_job

    def run_bulk_jobs(self, template, begin_index, end_index, step, max_parallel, auth=None):
        """ 
        Run an array job using template. 

        :param template: Job template; it can be specified either as a dictionary, or directly as a JobTemplate object.
        :type template: JobTemplate or dict

        :param begin_index: Begin index.
        :type begin_index: long

        :param end_index: End index.
        :type end_index: long

        :param step: Index step. 
        :type step: long

        :param max_parallel: Max. number of parallel tasks.
        :type max_parallel: long

        :param auth: Optional sudo object for running the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        :returns: JobArray object.

        >>> ja = j_session.run_bulk_jobs({'remote_command' : '/bin/sleep', 'args' : ['10']}, 1, 10, 3, 2)
        >>> print(ja)
        {'id': '528', 'job_list': [Job({'id': '528.1', 'session_name': 'js-01'}), Job({'id': '528.4', 'session_name': 'js-01'}), Job({'id': '528.7', 'session_name': 'js-01'}), Job({'id': '528.10', 'session_name': 'js-01'})], 'session_name': 'js-01'}
        >>> type(ja)
        <class 'drmaa2.job_array.JobArray'>
        """
        self.logger.debug(
            'Running an array job with task indices ({},{},{}), max. parallel {}, and using template: {}'.format(
                begin_index, end_index, step, max_parallel, template))
        drmaa2_lib = self.get_drmaa2_library()
        template = JobTemplate.create_from_dict(template)

        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ctypes_job_array = drmaa2_lib.drmaa2_jsession_run_bulk_jobs_as(auth._struct, self._struct, template._struct,
                                                                           begin_index, end_index, step, max_parallel)
        else:
            ctypes_job_array = drmaa2_lib.drmaa2_jsession_run_bulk_jobs(self._struct, template._struct, begin_index,
                                                                        end_index, step, max_parallel)
        if not ctypes_job_array:
            self.exception_mapper.check_last_error_code()
        py_job_array = JobArray(ctypes_job_array)
        self.logger.debug('Got job array {}'.format(py_job_array))
        return py_job_array

    def wait_any_started(self, job_list, timeout=Time.INFINITE_TIME):
        """ 
        Wait for start of any job from the given list . 

        :param job_list: Job list.
        :type job_list: [Job]

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        :returns: Started Job object.

        >>> ja = j_session.run_bulk_jobs({'remote_command' : '/bin/sleep', 'args' : ['10']}, 1, 10, 3, 2)
        >>> print(ja)
        {'id': '542', 'job_list': [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})], 'session_name': 'js-01'}
        >>> j = j_session.wait_any_started(ja.job_list)
        >>> print(j)
        {'id': '542.1', 'session_name': 'js-01'}
        """
        self.logger.debug('Wating for any one of {} jobs to start'.format(len(job_list)))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_list = Job.to_ctypes_job_list(job_list)
        ctypes_job = drmaa2_lib.drmaa2_jsession_wait_any_started(self._struct, ctypes_job_list, int(timeout))
        if not ctypes_job:
            self.exception_mapper.check_last_error_code()
        py_job = Job(ctypes_job)
        self.logger.debug('Job {} started'.format(py_job))
        drmaa2_lib.drmaa2_j_free(pointer(ctypes_job))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        return py_job

    def wait_any_terminated(self, job_list, timeout=Time.INFINITE_TIME):
        """ 
        Wait for termination of any job from the given list . 

        :param job_list: Job list.
        :type job_list: [Job]

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        :returns: Terminated Job object.

        >>> ja = j_session.run_bulk_jobs({'remote_command' : '/bin/sleep', 'args' : ['10']}, 1, 10, 3, 2)
        >>> print(ja)
        {'id': '542', 'job_list': [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})], 'session_name': 'js-01'}
        >>> j = j_session.wait_any_terminated(ja.job_list)
        >>> print(j)
        {'id': '542.1', 'session_name': 'js-01'}
        """
        self.logger.debug('Wating for any one of {} jobs to terminate'.format(len(job_list)))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_list = Job.to_ctypes_job_list(job_list)
        ctypes_job = drmaa2_lib.drmaa2_jsession_wait_any_terminated(self._struct, ctypes_job_list, int(timeout))
        if not ctypes_job:
            self.exception_mapper.check_last_error_code()
        py_job = Job(ctypes_job)
        self.logger.debug('Job {} terminated'.format(py_job))
        drmaa2_lib.drmaa2_j_free(pointer(ctypes_job))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        return py_job

    def wait_all_started(self, job_list, timeout=Time.INFINITE_TIME):
        """ 
        Wait for start of all jobs from the given list . 

        :param job_list: Job list.
        :type job_list: [Job]

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        :returns: List of started Job objects.

        >>> ja = j_session.run_bulk_jobs({'remote_command' : '/bin/sleep', 'args' : ['10']}, 1, 10, 3, 2)
        >>> print(ja)
        {'id': '542', 'job_list': [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})], 'session_name': 'js-01'}
        >>> jl = js.wait_all_started(ja.job_list)
        >>> print(jl)
        [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})]
        """
        self.logger.debug('Wating for all of {} jobs to start'.format(len(job_list)))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_list = Job.to_ctypes_job_list(job_list)
        ctypes_job_list2 = drmaa2_lib.drmaa2_jsession_wait_all_started(self._struct, ctypes_job_list, int(timeout))
        if not ctypes_job_list2:
            self.exception_mapper.check_last_error_code()
        py_job_list = Job.to_py_job_list(ctypes_job_list2)
        self.logger.debug('All {} jobs started'.format(len(job_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list2)))
        return py_job_list

    def wait_all_terminated(self, job_list, timeout=Time.INFINITE_TIME):
        """ 
        Wait for termination of all jobs from the given list . 

        :param job_list: Job list.
        :type job_list: [Job]

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        :returns: List of terminated Job objects.

        >>> ja = j_session.run_bulk_jobs({'remote_command' : '/bin/sleep', 'args' : ['10']}, 1, 10, 3, 2)
        >>> print(ja)
        {'id': '542', 'job_list': [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})], 'session_name': 'js-01'}
        >>> jl = js.wait_all_terminated(ja.job_list)
        >>> print(jl)
        [Job({'id': '542.1', 'session_name': 'js-01'}), Job({'id': '542.4', 'session_name': 'js-01'}), Job({'id': '542.7', 'session_name': 'js-01'}), Job({'id': '542.10', 'session_name': 'js-01'})]
        """
        self.logger.debug('Wating for all of {} jobs to terminate'.format(len(job_list)))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_list = Job.to_ctypes_job_list(job_list)
        ctypes_job_list2 = drmaa2_lib.drmaa2_jsession_wait_all_terminated(self._struct, ctypes_job_list, int(timeout))
        if not ctypes_job_list2:
            self.exception_mapper.check_last_error_code()
        py_job_list = Job.to_py_job_list(ctypes_job_list2)
        self.logger.debug('All {} jobs terminated'.format(len(job_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list2)))
        return py_job_list

    def get_job_array(self, id):
        """ 
        Get job array.

        :param id: Job array id.
        :type id: str

        :returns: Job array object.

        >>> ja = j_session.get_job_array(ja_id)
        """
        self.logger.debug('Retrieving job array id {}'.format(id))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_array = drmaa2_lib.drmaa2_jsession_get_job_array(self._struct, ByteString(id).encode())
        if not ctypes_job_array:
            self.exception_mapper.check_last_error_code()
        py_job_array = JobArray(ctypes_job_array)
        self.logger.debug('Got job array {}'.format(py_job_array))
        return py_job_array

    def get_job_categories(self):
        """ 
        Get list of all job categories.

        :returns: List of job categories

        >>> j_session = JobSession('js-01')
        >>> print(j_session.get_job_categories())
        []
        """
        drmaa2_lib = self.get_drmaa2_library()
        return Drmaa2Object.to_py_string_list(drmaa2_lib.drmaa2_jsession_get_job_categories(self._struct))

    def get_jobs(self, filter):
        """ 
        Get jobs matching the specified info.

        :param filter: Job info filter.
        :type filter: JobInfo

        :returns: List of Job objects.

        >>> j_info = JobInfo({'job_name' : 'a_job'})
        >>> j_list = j_session.get_all_jobs(j_info)
        """
        self.logger.debug('Requesting list of jobs using filter: {}'.format(filter))
        drmaa2_lib = self.get_drmaa2_library()
        job_info = filter
        if type(filter) == PY_DICT_TYPE:
            job_info = JobInfo(filter)
        ctypes_filter = job_info._struct
        ctypes_job_list = drmaa2_lib.drmaa2_jsession_get_jobs(self._struct, ctypes_filter)
        if not ctypes_job_list:
            self.exception_mapper.check_last_error_code()

        py_job_list = []
        if ctypes_job_list:
            py_job_list = Job.to_py_job_list(ctypes_job_list)
            drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        return py_job_list



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

class Singleton(object):
    """ Base class for singleton objects. """

    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None or cls != type(cls.__instance):
            instance = object.__new__(cls, *args, **kwargs)
            instance.__init__()
            cls.__instance = instance
        return cls.__instance

    @classmethod
    def get_instance(cls, *args, **kwargs):
        """ 
        Get singleton instance.

        >>> s = Singleton.get_instance()
        """
        return cls.__new__(cls, *args, **kwargs)

    def __init__(self):
        if self.__instance is not None:
            return


#######################################################################
# Test.
if __name__ == '__main__':
    a = Singleton()
    print('A=%s' % a)
    b = Singleton()
    print('B=%s' % b)
    c = Singleton.get_instance()
    print('C=%s' % c)



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

import os
import glob
import platform
import ctypes
import ctypes.util
from ctypes import c_char_p
from ctypes import c_void_p
from ctypes import c_long
from ctypes import c_longlong
from ctypes import POINTER
from ctypes import pointer

from .drmaa2_ctypes import drmaa2_error
from .drmaa2_ctypes import drmaa2_bool
from .drmaa2_ctypes import drmaa2_capability

from .drmaa2_ctypes import drmaa2_string
from .drmaa2_ctypes import drmaa2_string_list
from .drmaa2_ctypes import drmaa2_list
from .drmaa2_ctypes import drmaa2_listtype
from .drmaa2_ctypes import drmaa2_list_entryfree
from .drmaa2_ctypes import drmaa2_dict
from .drmaa2_ctypes import drmaa2_dict_entryfree

from .drmaa2_ctypes import drmaa2_j
from .drmaa2_ctypes import drmaa2_j_list
from .drmaa2_ctypes import drmaa2_jarray
from .drmaa2_ctypes import drmaa2_jinfo
from .drmaa2_ctypes import drmaa2_jtemplate
from .drmaa2_ctypes import drmaa2_jstate

from .drmaa2_ctypes import drmaa2_r
from .drmaa2_ctypes import drmaa2_r_list
from .drmaa2_ctypes import drmaa2_rinfo
from .drmaa2_ctypes import drmaa2_rtemplate

from .drmaa2_ctypes import drmaa2_slotinfo
from .drmaa2_ctypes import drmaa2_queueinfo
from .drmaa2_ctypes import drmaa2_queueinfo_list

from .drmaa2_ctypes import drmaa2_machineinfo
from .drmaa2_ctypes import drmaa2_machineinfo_list

from .drmaa2_ctypes import drmaa2_jsession
from .drmaa2_ctypes import drmaa2_rsession
from .drmaa2_ctypes import drmaa2_msession

from .drmaa2_ctypes import drmaa2_sudo
from .drmaa2_ctypes import drmaa2_notification
from .drmaa2_ctypes import drmaa2_version
from .drmaa2_ctypes import drmaa2_time
from .drmaa2_ctypes import drmaa2_callback

from .byte_string import ByteString
from .log_manager import LogManager
from .singleton import Singleton
from .drmaa2_exceptions import Drmaa2Exception


class LibraryManager(Singleton):
    """ 
    Singleton class for loading and keeping reference to the
    underlying C library.
    """

    logger = LogManager.get_instance().get_logger('LibraryManager')

    __instance = None

    def __init__(self):
        """
        Constructor. 

        >>> lm = LibraryManager()
        """
        if LibraryManager.__instance:
            return
        LibraryManager.__instance = self
        self.drmaa2_library = self.__load_drmaa2_library()

    def get_drmaa2_library(self):
        """
        Get reference to the DRMAA2 C library.

        >>> drmaa2_lib = LibraryManager.get_instance().get_drmaa2_library()
        """
        return self.drmaa2_library

    def to_py_string(self, ctypes_string):
        py_string = ByteString(ctypes_string.value).decode()
        self.drmaa2_library.drmaa2_string_free(pointer(ctypes_string))
        return py_string

    @classmethod
    def get_drms_name(cls):
        """
        Retrieve DRMS name.

        :returns: DRMS name

        >>> print(LibraryManager.get_drms_name())
        Univa Grid Engine
        """
        lm = LibraryManager.get_instance()
        ctypes_string = lm.get_drmaa2_library().drmaa2_get_drms_name();
        return lm.to_py_string(ctypes_string)

    @classmethod
    def get_drmaa_name(cls):
        """
        Retrieve DRMAA name.

        :returns: DRMAA name

        >>> print(LibraryManager.get_drmaa_name())
        Univa Grid Engine Drmaa V2
        """
        lm = LibraryManager.get_instance()
        ctypes_string = lm.get_drmaa2_library().drmaa2_get_drmaa_name();
        return lm.to_py_string(ctypes_string)

    @classmethod
    def drmaa_supports(cls, capability):
        """
        Check whether the library supports given capability/

        :returns: True if capability is supported, false otherwise.

        >>> print(LibraryManager.drmaa_supports(Capability.JT_EMAIL))
        True
        """
        lm = LibraryManager.get_instance()
        c = capability
        return lm.get_drmaa2_library().drmaa2_supports(int(c)) > 0;

    @classmethod
    def __load_drmaa2_library(cls):
        cls.logger.debug('Loading DRMAA2 library')
        libc_name = ctypes.util.find_library('c')
        libc = ctypes.CDLL(libc_name)
        libc.memcmp.argtypes = (ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)
        lib_path = 'libdrmaa2.so'
        drmaa2_lib = None
        try:
            SGE_ROOT = os.environ['SGE_ROOT']
            p = os.popen(SGE_ROOT + '/util/arch')
            try:
                SGE_ARCH = p.read().rstrip()
            finally:
                p.close()
            lib_dir = SGE_ROOT + '/drmaa/lib/' + SGE_ARCH
            cls.logger.debug('Looking for DRMAA2 library under %s' % lib_dir)
            lib_paths = glob.glob(lib_dir + '/libdrmaa2.so')
            if not lib_paths:
                lib_paths = glob.glob(lib_dir + '/libdrmaa2.dylib')
            if len(lib_paths):
                lib_path = lib_paths[0]
            else:
                cls.logger.warn('Could not find DRMAA2 library under %s' % lib_dir)
        except KeyError:
            cls.logger.debug('SGE_ROOT is not defined')
        cls.logger.debug('Library path: %s' % lib_path)

        try:
            drmaa2_lib = ctypes.cdll.LoadLibrary(str(lib_path))
        except OSError:
            raise Drmaa2Exception('Could not load DRMAA2 library.')

        cls.logger.debug("Initializing DRMAA2 library")

        drmaa2_lib.drmaa2_string_free.restype = None
        drmaa2_lib.drmaa2_string_free.argtypes = [POINTER(drmaa2_string)]

        drmaa2_lib.drmaa2_list_create.restype = drmaa2_list
        drmaa2_lib.drmaa2_list_create.argtypes = [drmaa2_listtype, drmaa2_list_entryfree]
        drmaa2_lib.drmaa2_list_free.restype = None
        drmaa2_lib.drmaa2_list_free.argtypes = [POINTER(drmaa2_list)]
        drmaa2_lib.drmaa2_list_get.restype = c_void_p
        drmaa2_lib.drmaa2_list_get.argtypes = [drmaa2_list, c_long]
        drmaa2_lib.drmaa2_list_add.restype = drmaa2_error
        drmaa2_lib.drmaa2_list_add.argtypes = [drmaa2_list, c_void_p]
        drmaa2_lib.drmaa2_list_del.restype = drmaa2_error
        drmaa2_lib.drmaa2_list_del.argtypes = [drmaa2_list, c_long]
        drmaa2_lib.drmaa2_list_size.restype = c_long
        drmaa2_lib.drmaa2_list_size.argtypes = [drmaa2_list]

        drmaa2_lib.drmaa2_lasterror.restype = drmaa2_error
        drmaa2_lib.drmaa2_lasterror.argtypes = []
        drmaa2_lib.drmaa2_lasterror_text.restype = drmaa2_string
        drmaa2_lib.drmaa2_lasterror_text.argtypes = []

        # UGE-specific
        drmaa2_lib.uge_drmaa2_list_free_root.restype = None
        drmaa2_lib.uge_drmaa2_list_free_root.argtypes = [POINTER(drmaa2_list)]
        drmaa2_lib.uge_drmaa2_list_set.restype = drmaa2_error
        drmaa2_lib.uge_drmaa2_list_set.argtypes = [drmaa2_list, c_long, c_void_p]

        drmaa2_lib.uge_vi_impl_spec_get.restype = drmaa2_dict
        drmaa2_lib.uge_vi_impl_spec_get.argtypes = [POINTER(drmaa2_version)]

        drmaa2_lib.drmaa2_dict_create.restype = drmaa2_dict
        drmaa2_lib.drmaa2_dict_create.argtypes = [drmaa2_dict_entryfree]
        drmaa2_lib.drmaa2_dict_free.restype = None
        drmaa2_lib.drmaa2_dict_free.argtypes = [POINTER(drmaa2_dict)]
        drmaa2_lib.drmaa2_dict_list.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_dict_list.argtypes = [drmaa2_dict]
        drmaa2_lib.drmaa2_dict_has.restype = drmaa2_bool
        drmaa2_lib.drmaa2_dict_has.argtypes = [drmaa2_dict, c_char_p]
        drmaa2_lib.drmaa2_dict_get.restype = c_char_p
        drmaa2_lib.drmaa2_dict_get.argtypes = [drmaa2_dict, c_char_p]
        drmaa2_lib.drmaa2_dict_del.restype = drmaa2_error
        drmaa2_lib.drmaa2_dict_del.argtypes = [drmaa2_dict, c_char_p]
        drmaa2_lib.drmaa2_dict_set.restype = drmaa2_error
        drmaa2_lib.drmaa2_dict_set.argtypes = [drmaa2_dict, c_char_p]

        drmaa2_lib.drmaa2_jinfo_create.restype = POINTER(drmaa2_jinfo)
        drmaa2_lib.drmaa2_jinfo_create.argtypes = []
        drmaa2_lib.drmaa2_jinfo_free.restype = None
        drmaa2_lib.drmaa2_jinfo_free.argtypes = [POINTER(POINTER(drmaa2_jinfo))]

        drmaa2_lib.drmaa2_slotinfo_free.restype = None
        drmaa2_lib.drmaa2_slotinfo_free.argtypes = [POINTER(POINTER(drmaa2_slotinfo))]

        drmaa2_lib.drmaa2_rinfo_create.restype = POINTER(drmaa2_rinfo)
        drmaa2_lib.drmaa2_rinfo_create.argtypes = []
        drmaa2_lib.drmaa2_rinfo_free.restype = None
        drmaa2_lib.drmaa2_rinfo_free.argtypes = [POINTER(POINTER(drmaa2_rinfo))]

        drmaa2_lib.drmaa2_jtemplate_create.restype = POINTER(drmaa2_jtemplate)
        drmaa2_lib.drmaa2_jtemplate_create.argtypes = []
        drmaa2_lib.drmaa2_jtemplate_free.restype = None
        drmaa2_lib.drmaa2_jtemplate_free.argtypes = [POINTER(POINTER(drmaa2_jtemplate))]

        drmaa2_lib.drmaa2_rtemplate_create.restype = POINTER(drmaa2_rtemplate)
        drmaa2_lib.drmaa2_rtemplate_create.argtypes = []
        drmaa2_lib.drmaa2_rtemplate_free.restype = None
        drmaa2_lib.drmaa2_rtemplate_free.argtypes = [POINTER(POINTER(drmaa2_rtemplate))]

        drmaa2_lib.drmaa2_queueinfo_free.restype = None
        drmaa2_lib.drmaa2_queueinfo_free.argtypes = [POINTER(POINTER(drmaa2_queueinfo))]

        drmaa2_lib.drmaa2_machineinfo_free.restype = None
        drmaa2_lib.drmaa2_machineinfo_free.argtypes = [POINTER(POINTER(drmaa2_machineinfo))]

        drmaa2_lib.drmaa2_notification_free.restype = None
        drmaa2_lib.drmaa2_notification_free.argtypes = [POINTER(POINTER(drmaa2_notification))]

        drmaa2_lib.drmaa2_version_free.restype = None
        drmaa2_lib.drmaa2_version_free.argtypes = [POINTER(POINTER(drmaa2_version))]

        drmaa2_lib.drmaa2_jtemplate_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_jtemplate_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_jinfo_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_jinfo_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_rtemplate_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_rtemplate_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_rinfo_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_rinfo_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_queueinfo_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_queueinfo_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_machineinfo_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_machineinfo_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_notification_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_notification_impl_spec.argtypes = []
        drmaa2_lib.drmaa2_version_impl_spec.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_version_impl_spec.argtypes = []

        drmaa2_lib.drmaa2_get_instance_value.restype = drmaa2_string
        drmaa2_lib.drmaa2_get_instance_value.argtypes = [c_void_p, c_char_p]
        drmaa2_lib.drmaa2_describe_attribute.restype = drmaa2_string
        drmaa2_lib.drmaa2_describe_attribute.argtypes = [c_void_p, c_char_p]
        drmaa2_lib.drmaa2_set_instance_value.restype = drmaa2_error
        drmaa2_lib.drmaa2_set_instance_value.argtypes = [c_void_p, c_char_p, c_char_p]

        drmaa2_lib.drmaa2_jsession_free.restype = None
        drmaa2_lib.drmaa2_jsession_free.argtypes = [POINTER(POINTER(drmaa2_jsession))]
        drmaa2_lib.drmaa2_rsession_free.restype = None
        drmaa2_lib.drmaa2_rsession_free.argtypes = [POINTER(POINTER(drmaa2_rsession))]
        drmaa2_lib.drmaa2_msession_free.restype = None
        drmaa2_lib.drmaa2_msession_free.argtypes = [POINTER(POINTER(drmaa2_msession))]

        drmaa2_lib.drmaa2_j_free.restype = None
        drmaa2_lib.drmaa2_j_free.argtypes = [POINTER(POINTER(drmaa2_j))]
        drmaa2_lib.drmaa2_jarray_free.restype = None
        drmaa2_lib.drmaa2_jarray_free.argtypes = [POINTER(POINTER(drmaa2_jarray))]
        drmaa2_lib.drmaa2_r_free.restype = None
        drmaa2_lib.drmaa2_r_free.argtypes = [POINTER(POINTER(drmaa2_r))]

        drmaa2_lib.drmaa2_rsession_get_contact.restype = drmaa2_string
        drmaa2_lib.drmaa2_rsession_get_contact.argtypes = [drmaa2_rsession]
        drmaa2_lib.drmaa2_rsession_get_session_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_rsession_get_session_name.argtypes = [drmaa2_rsession]
        drmaa2_lib.drmaa2_rsession_get_reservation.restype = POINTER(drmaa2_r)
        drmaa2_lib.drmaa2_rsession_get_reservation.argtypes = [POINTER(drmaa2_rsession), drmaa2_string]
        drmaa2_lib.drmaa2_rsession_request_reservation.restype = POINTER(drmaa2_r)
        drmaa2_lib.drmaa2_rsession_request_reservation.argtypes = [POINTER(drmaa2_rsession), POINTER(drmaa2_rtemplate)]
        drmaa2_lib.drmaa2_rsession_request_reservation_as.restype = POINTER(drmaa2_r)
        drmaa2_lib.drmaa2_rsession_request_reservation_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_rsession),
                                                                      POINTER(drmaa2_rtemplate)]

        drmaa2_lib.drmaa2_rsession_get_reservations.restype = drmaa2_r_list
        drmaa2_lib.drmaa2_rsession_get_reservations.argtypes = [POINTER(drmaa2_rsession)]

        drmaa2_lib.drmaa2_r_get_id.restype = drmaa2_string
        drmaa2_lib.drmaa2_r_get_id.argtypes = [drmaa2_r]
        drmaa2_lib.drmaa2_r_get_session_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_r_get_session_name.argtypes = [drmaa2_r]
        # drmaa2_lib.drmaa2_r_get_reservation_template.restype = POINTER(drmaa2_rtemplate)
        # drmaa2_lib.drmaa2_r_get_reservation_template.argtypes = [POINTER(drmaa2_r)]
        drmaa2_lib.drmaa2_r_get_rtemplate.restype = POINTER(drmaa2_rtemplate)
        drmaa2_lib.drmaa2_r_get_rtemplate.argtypes = [POINTER(drmaa2_r)]
        drmaa2_lib.drmaa2_r_get_info.restype = POINTER(drmaa2_rinfo)
        drmaa2_lib.drmaa2_r_get_info.argtypes = [POINTER(drmaa2_r)]
        drmaa2_lib.drmaa2_r_terminate.restype = drmaa2_error
        drmaa2_lib.drmaa2_r_terminate.argtypes = [POINTER(drmaa2_r)]
        drmaa2_lib.drmaa2_r_terminate_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_r_terminate_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_r)]

        drmaa2_lib.drmaa2_jarray_get_id.restype = drmaa2_string
        drmaa2_lib.drmaa2_jarray_get_id.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_get_jobs.restype = drmaa2_j_list
        drmaa2_lib.drmaa2_jarray_get_jobs.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_get_session_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_jarray_get_session_name.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_get_jtemplate.restype = POINTER(drmaa2_jtemplate)
        drmaa2_lib.drmaa2_jarray_get_jtemplate.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_suspend.restype = drmaa2_error
        drmaa2_lib.drmaa2_jarray_suspend.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_resume.restype = drmaa2_error
        drmaa2_lib.drmaa2_jarray_resume.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_hold.restype = drmaa2_error
        drmaa2_lib.drmaa2_jarray_hold.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_release.restype = drmaa2_error
        drmaa2_lib.drmaa2_jarray_release.argtypes = [POINTER(drmaa2_jarray)]
        drmaa2_lib.drmaa2_jarray_terminate.restype = drmaa2_error
        drmaa2_lib.drmaa2_jarray_terminate.argtypes = [POINTER(drmaa2_jarray)]

        drmaa2_lib.drmaa2_jsession_get_contact.restype = drmaa2_string
        drmaa2_lib.drmaa2_jsession_get_contact.argtypes = [POINTER(drmaa2_jsession)]
        drmaa2_lib.drmaa2_jsession_get_session_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_jsession_get_session_name.argtypes = [POINTER(drmaa2_jsession)]
        drmaa2_lib.drmaa2_jsession_get_job_categories.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_jsession_get_job_categories.argtypes = [POINTER(drmaa2_jsession)]
        drmaa2_lib.drmaa2_jsession_get_jobs.restype = drmaa2_j_list
        drmaa2_lib.drmaa2_jsession_get_jobs.argtypes = [POINTER(drmaa2_jsession), POINTER(drmaa2_jinfo)]
        drmaa2_lib.drmaa2_jsession_get_job_array.restype = POINTER(drmaa2_jarray)
        drmaa2_lib.drmaa2_jsession_get_job_array.argtypes = [POINTER(drmaa2_jsession), drmaa2_string]
        drmaa2_lib.drmaa2_jsession_run_job.restype = POINTER(drmaa2_j)
        drmaa2_lib.drmaa2_jsession_run_job.argtypes = [POINTER(drmaa2_jsession), POINTER(drmaa2_jtemplate)]
        drmaa2_lib.drmaa2_jsession_run_job_as.restype = POINTER(drmaa2_j)
        drmaa2_lib.drmaa2_jsession_run_job_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_jsession),
                                                          POINTER(drmaa2_jtemplate)]
        drmaa2_lib.drmaa2_jsession_run_bulk_jobs.restype = POINTER(drmaa2_jarray)
        drmaa2_lib.drmaa2_jsession_run_bulk_jobs.argtypes = [POINTER(drmaa2_jsession), POINTER(drmaa2_jtemplate),
                                                             c_longlong, c_longlong, c_longlong, c_longlong]
        drmaa2_lib.drmaa2_jsession_wait_any_started.restype = POINTER(drmaa2_j)
        drmaa2_lib.drmaa2_jsession_wait_any_started.argtypes = [POINTER(drmaa2_jsession), drmaa2_j_list, drmaa2_time]
        drmaa2_lib.drmaa2_jsession_wait_any_terminated.restype = POINTER(drmaa2_j)
        drmaa2_lib.drmaa2_jsession_wait_any_terminated.argtypes = [POINTER(drmaa2_jsession), drmaa2_j_list, drmaa2_time]
        drmaa2_lib.drmaa2_jsession_wait_all_started.restype = drmaa2_j_list
        drmaa2_lib.drmaa2_jsession_wait_all_started.argtypes = [POINTER(drmaa2_jsession), drmaa2_j_list, drmaa2_time]
        drmaa2_lib.drmaa2_jsession_wait_all_terminated.restype = drmaa2_j_list
        drmaa2_lib.drmaa2_jsession_wait_all_terminated.argtypes = [POINTER(drmaa2_jsession), drmaa2_j_list, drmaa2_time]
        drmaa2_lib.drmaa2_j_suspend.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_suspend.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_suspend_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_suspend_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_resume.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_resume.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_resume_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_resume_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_hold.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_hold.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_hold_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_hold_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_release.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_release.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_release_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_release_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_terminate.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_terminate_forced.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate_forced.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_terminate_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j), drmaa2_bool]
        drmaa2_lib.drmaa2_j_terminate_all.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate_all.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_terminate_forced_all.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate_forced_all.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_terminate_all_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_terminate_all_as.argtypes = [POINTER(drmaa2_sudo), POINTER(drmaa2_j), drmaa2_bool]
        drmaa2_lib.drmaa2_j_reap.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_reap.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_get_id.restype = drmaa2_string
        drmaa2_lib.drmaa2_j_get_id.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_get_jtemplate.restype = POINTER(drmaa2_jtemplate)
        drmaa2_lib.drmaa2_j_get_jtemplate.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_get_state.restype = drmaa2_jstate
        drmaa2_lib.drmaa2_j_get_state.argtypes = [POINTER(drmaa2_j), POINTER(drmaa2_string)]
        drmaa2_lib.drmaa2_j_get_info.restype = POINTER(drmaa2_jinfo)
        drmaa2_lib.drmaa2_j_get_info.argtypes = [POINTER(drmaa2_j)]
        drmaa2_lib.drmaa2_j_wait_started.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_wait_started.argtypes = [POINTER(drmaa2_j), drmaa2_time]
        drmaa2_lib.drmaa2_j_wait_terminated.restype = drmaa2_error
        drmaa2_lib.drmaa2_j_wait_terminated.argtypes = [POINTER(drmaa2_j)]

        drmaa2_lib.drmaa2_msession_get_all_reservations.restype = drmaa2_r_list
        drmaa2_lib.drmaa2_msession_get_all_reservations.argtypes = [POINTER(drmaa2_msession)]

        drmaa2_lib.drmaa2_msession_get_all_jobs.restype = drmaa2_j_list
        drmaa2_lib.drmaa2_msession_get_all_jobs.argtypes = [POINTER(drmaa2_msession), POINTER(drmaa2_jinfo)]
        drmaa2_lib.drmaa2_msession_get_all_queues.restype = drmaa2_queueinfo_list
        drmaa2_lib.drmaa2_msession_get_all_queues.argtypes = [POINTER(drmaa2_msession), drmaa2_string_list]
        drmaa2_lib.drmaa2_msession_get_all_machines.restype = drmaa2_machineinfo_list
        drmaa2_lib.drmaa2_msession_get_all_machines.argtypes = [POINTER(drmaa2_msession), drmaa2_string_list]

        drmaa2_lib.drmaa2_get_drms_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_get_drms_name.argtypes = []
        drmaa2_lib.drmaa2_get_drms_version.restype = POINTER(drmaa2_version)
        drmaa2_lib.drmaa2_get_drms_version.argtypes = []
        drmaa2_lib.drmaa2_get_drmaa_name.restype = drmaa2_string
        drmaa2_lib.drmaa2_get_drmaa_name.argtypes = []
        drmaa2_lib.drmaa2_get_drmaa_version.restype = POINTER(drmaa2_version)
        drmaa2_lib.drmaa2_get_drmaa_version.argtypes = []
        drmaa2_lib.drmaa2_supports.restype = drmaa2_bool
        drmaa2_lib.drmaa2_supports.argtypes = [drmaa2_capability]
        drmaa2_lib.drmaa2_create_jsession.restype = POINTER(drmaa2_jsession)
        drmaa2_lib.drmaa2_create_jsession.argtypes = [c_char_p, c_char_p]
        drmaa2_lib.drmaa2_create_jsession_as.restype = POINTER(drmaa2_jsession)
        drmaa2_lib.drmaa2_create_jsession_as.argtypes = [POINTER(drmaa2_sudo), c_char_p, c_char_p]
        drmaa2_lib.drmaa2_create_rsession.restype = POINTER(drmaa2_rsession)
        drmaa2_lib.drmaa2_create_rsession.argtypes = [c_char_p, c_char_p]
        drmaa2_lib.drmaa2_create_rsession_as.restype = POINTER(drmaa2_rsession)
        drmaa2_lib.drmaa2_create_rsession_as.argtypes = [POINTER(drmaa2_sudo), c_char_p, c_char_p]
        drmaa2_lib.drmaa2_open_jsession.restype = POINTER(drmaa2_jsession)
        drmaa2_lib.drmaa2_open_jsession.argtypes = [c_char_p]
        drmaa2_lib.drmaa2_open_rsession.restype = POINTER(drmaa2_rsession)
        drmaa2_lib.drmaa2_open_rsession.argtypes = [c_char_p]
        drmaa2_lib.drmaa2_open_msession.restype = POINTER(drmaa2_msession)
        drmaa2_lib.drmaa2_open_msession.argtypes = [c_char_p]
        drmaa2_lib.drmaa2_close_jsession.restype = drmaa2_error
        drmaa2_lib.drmaa2_close_jsession.argtypes = [POINTER(drmaa2_jsession)]
        drmaa2_lib.drmaa2_close_rsession.restype = drmaa2_error
        drmaa2_lib.drmaa2_close_rsession.argtypes = [POINTER(drmaa2_rsession)]
        drmaa2_lib.drmaa2_close_msession.restype = drmaa2_error
        drmaa2_lib.drmaa2_close_msession.argtypes = [POINTER(drmaa2_msession)]
        drmaa2_lib.drmaa2_destroy_jsession.restype = drmaa2_error
        drmaa2_lib.drmaa2_destroy_jsession.argtypes = [c_char_p]
        drmaa2_lib.drmaa2_destroy_jsession_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_destroy_jsession_as.argtypes = [POINTER(drmaa2_sudo), c_char_p]
        drmaa2_lib.drmaa2_destroy_rsession.restype = drmaa2_error
        drmaa2_lib.drmaa2_destroy_rsession.argtypes = [c_char_p]
        drmaa2_lib.drmaa2_destroy_rsession_as.restype = drmaa2_error
        drmaa2_lib.drmaa2_destroy_rsession_as.argtypes = [POINTER(drmaa2_sudo), c_char_p]
        drmaa2_lib.drmaa2_get_jsession_names.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_get_jsession_names.argtypes = []
        drmaa2_lib.drmaa2_get_rsession_names.restype = drmaa2_string_list
        drmaa2_lib.drmaa2_get_rsession_names.argtypes = []
        drmaa2_lib.drmaa2_register_event_notification.restype = drmaa2_error
        drmaa2_lib.drmaa2_register_event_notification.argtypes = [POINTER(drmaa2_callback)]
        cls.logger.debug("Done initializing DRMAA2 library")
        return drmaa2_lib


#######################################################################
# Test.
if __name__ == '__main__':
    lm = LibraryManager()
    print(lm.get_drmaa2_library())



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

"""
Exception mapper class.
"""

from .byte_string import ByteString
from .drmaa2_constants import StatusCode
from .drmaa2_exceptions import Drmaa2Exception
from .drmaa2_exceptions import DeniedByDrms
from .drmaa2_exceptions import DrmCommunicationError
from .drmaa2_exceptions import TryLaterError
from .drmaa2_exceptions import SessionManagementError
from .drmaa2_exceptions import TimeoutError
from .drmaa2_exceptions import InternalError
from .drmaa2_exceptions import InvalidArgument
from .drmaa2_exceptions import InvalidSession
from .drmaa2_exceptions import InvalidState
from .drmaa2_exceptions import ResourceNotAvailable
from .drmaa2_exceptions import UnsupportedAttribute
from .drmaa2_exceptions import UnsupportedOperation
from .drmaa2_exceptions import ImplementationSpecificError
from .drmaa2_exceptions import AuthorizationError
from .library_manager import LibraryManager


class ExceptionMapper(object):
    EXCEPTION_MAP = {
        StatusCode.UNSET_ERROR: Drmaa2Exception,
        StatusCode.SUCCESS: None,
        StatusCode.DENIED_BY_DRMS: DeniedByDrms,
        StatusCode.DRM_COMMUNICATION: DrmCommunicationError,
        StatusCode.TRY_LATER: TryLaterError,
        StatusCode.SESSION_MANAGEMENT: SessionManagementError,
        StatusCode.TIMEOUT: TimeoutError,
        StatusCode.INTERNAL: InternalError,
        StatusCode.INVALID_ARGUMENT: InvalidArgument,
        StatusCode.INVALID_SESSION: InvalidSession,
        StatusCode.INVALID_STATE: InvalidState,
        StatusCode.OUT_OF_RESOURCE: ResourceNotAvailable,
        StatusCode.UNSUPPORTED_ATTRIBUTE: UnsupportedAttribute,
        StatusCode.UNSUPPORTED_OPERATION: UnsupportedOperation,
        StatusCode.IMPLEMENTATION_SPECIFIC: ImplementationSpecificError,
        StatusCode.AUTHORIZATION: AuthorizationError,
    }

    @classmethod
    def get_last_error_message(cls):
        """ 
        Retrieve the last error from the DRMAA2 library.

        :return str: Last error message
        """
        drmaa2_lib = LibraryManager.get_instance().get_drmaa2_library()
        string_ptr = drmaa2_lib.drmaa2_lasterror_text()
        message = ''
        if string_ptr:
            message = ByteString(string_ptr.value).decode()
            drmaa2_lib.drmaa2_string_free(string_ptr)
        return message

    @classmethod
    def last_error_code(cls):
        """ 
        Retrieve the last error code from the DRMAA2 library.

        :return int: Last error number
        """
        drmaa2_lib = LibraryManager.get_instance().get_drmaa2_library()
        return drmaa2_lib.drmaa2_lasterror()

    @classmethod
    def check_last_error_code(cls):
        """ 
        Check the last error code from the DRMAA2 library.

        :raises Drmaa2Exception: in case of a non-zero code
        """
        code = cls.last_error_code()
        cls.check_status_code(code)

    @classmethod
    def check_status_code(cls, code):
        """ 
        Check the last error code from the DRMAA2 library.

        :raises Drmaa2Exception: in case of a non-zero code
        """
        status_code = StatusCode(code)
        if status_code != StatusCode.SUCCESS:
            error_class = cls.EXCEPTION_MAP.get(status_code, Drmaa2Exception)
            error = error_class(error=cls.get_last_error_message())
            raise error


#######################################################################
# Test.
if __name__ == '__main__':
    ExceptionMapper.check_last_error_code()
    ExceptionMapper.check_status_code(5)



#!/usr/bin/env python 
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from uuid import uuid4
from ctypes import pointer
from ctypes import c_void_p

from .drmaa2_constants import PY_DICT_TYPE

from .byte_string import ByteString
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_ctypes import drmaa2_machineinfo_list
from .drmaa2_object import Drmaa2Object

from .machine_info import MachineInfo
from .queue_info import QueueInfo
from .reservation_info import ReservationInfo
from .reservation import Reservation
from .job_info import JobInfo
from .job import Job
from .log_manager import LogManager
from .exception_mapper import ExceptionMapper


class MonitoringSession(Drmaa2Object):
    """ High-level DRMAA2 monitoring session class. """

    name = Drmaa2Object.StringDescriptor('name')
    """ Monitoring session name (str). """

    logger = LogManager.get_instance().get_logger('JobSession')
    exception_mapper = ExceptionMapper()

    def __init__(self, name=None, session_dict={}):
        """ 
        Constructor. If the session with a given name does not already exist, 
        it will create a new one; otherwise, it will open the existing session.

        :param name: Session name; it may be also be specified as part of the monitoring session object dictionary. If not provided, a random name will be used.
        :type name: str

        :param session_dict: Monitoring session dictionary; if specified, it should contain the 'name' key.
        :type session_dict: dict

        >>> m_session = MonitoringSession()
        >>> print(m_session.name)
        b6db641c6bf34ffaafaac15d31554778
        """
        Drmaa2Object.__init__(self)
        name = name or session_dict.get('name')
        if not name:
            name = uuid4().hex
        self._struct = self.__open(name)
        self._struct_p = pointer(self._struct)
        self._name_bs = ByteString(name)
        self._read_only = True

    def __open(self, name):
        self.logger.debug('Opening monitoring session {}'.format(name))
        struct = self.get_drmaa2_library().drmaa2_open_msession(ByteString(name).encode())
        if not struct:
            self.exception_mapper.check_last_error_code()
        return struct

    def open(self):
        """ 
        Open session. This method is called automatically in the constructor.

        >>> ...
        >>> m_session.close()
        >>> ...
        >>> m_session.open()
        >>> mi_list = m_session.get_all_machines(['univa.example.com'])
        """
        if self._struct:
            self.logger.debug('Monitoring session {} is already open'.format(self._name_bs.decode()))
        else:
            self._struct = self.__open(self._name_bs)
            self._struct_p = pointer(self._struct)

    def close(self):
        """ 
        Close session. This method is called automatically when the session
        object goes out of scope.

        >>> mi_list = m_session.get_all_machines(['univa.example.com'])
        >>> m_session.close()
        """
        if self._struct:
            self.logger.debug('Closing monitoring session {}'.format(self._name_bs.decode()))
            drmaa2_lib = self.get_drmaa2_library()
            self.exception_mapper.check_status_code(drmaa2_lib.drmaa2_close_msession(self._struct))
            drmaa2_lib.drmaa2_msession_free(self._struct_p)
            self._struct = None

    def __del__(self):
        """ Destructor. """
        self.close()

    def get_all_machines(self, filter):
        """ 
        Get information about specified machines.

        :param filter: List of machine names to retrieve information for.
        :type filter: [str]

        :returns: List of MachineInfo objects.

        >>> mi_list = m_session.get_all_machines(['univa.example.com'])
        >>> print(mi_list)
         [MachineInfo({'available': 'TRUE', 'cores_per_socket': 1, 'implementation_specific': {}, 'load': 0.07999999821186066, 'machine_arch': 'X64', 'machine_os': 'OTHER_OS', 'machine_os_version': Version({'implementation_specific': {}}), 'name': 'univ.example.com', 'phys_memory': 4047372, 'sockets': 1, 'threads_per_core': 1, 'virt_memory': 8110599})]
        """
        self.logger.debug('Requesting list of machines using filter: {}'.format(filter))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_filter = self.to_ctypes_string_list(filter)
        ctypes_machine_info_list = drmaa2_lib.drmaa2_msession_get_all_machines(self._struct, ctypes_filter)
        if not ctypes_machine_info_list:
            self.exception_mapper.check_last_error_code()

        py_machine_info_list = []
        if ctypes_machine_info_list:
            py_machine_info_list = MachineInfo.to_py_machine_info_list(ctypes_machine_info_list)
            drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_machine_info_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_filter)))
        return py_machine_info_list

    def get_all_queues(self, filter):
        """ 
        Get information about specified queues.

        :param filter: List of queue names to retrieve information for.
        :type filter: [str]

        :returns: List of QueueInfo objects.

        >>> qi_list = m_session.get_all_queues(['all.q'])
        >>> print(qi_list)
        [QueueInfo({'implementation_specific': {}, 'name': 'all.q'})]
        """
        self.logger.debug('Requesting list of queues using filter: {}'.format(filter))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_filter = self.to_ctypes_string_list(filter)
        ctypes_queue_info_list = drmaa2_lib.drmaa2_msession_get_all_queues(self._struct, ctypes_filter)
        if not ctypes_queue_info_list:
            self.exception_mapper.check_last_error_code()

        py_queue_info_list = []
        if ctypes_queue_info_list:
            py_queue_info_list = QueueInfo.to_py_queue_info_list(ctypes_queue_info_list)
            drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_queue_info_list)))
        drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_filter)))
        return py_queue_info_list

    def get_all_reservations(self, filter):
        """ 
        Get reservations matching the specified info.

        :param filter: Reservation info filter.
        :type filter: ReservationInfo

        :returns: List of Reservation objects.

        >>> ri = ReservationInfo({'reservation_name' : r_name})
        >>> r_list = m_session.get_all_reservations(ri)
        >>> print(r_list)
        [Reservation({'id': '49'})]
        """
        self.logger.debug('Requesting list of reservations using filter: {}'.format(filter))
        drmaa2_lib = self.get_drmaa2_library()
        reservation_info = filter
        if type(filter) == PY_DICT_TYPE:
            reservation_info = ReservationInfo(filter)
        ctypes_filter = reservation_info._struct
        ctypes_reservation_list = drmaa2_lib.drmaa2_msession_get_all_reservations(self._struct, ctypes_filter)
        if not ctypes_reservation_list:
            self.exception_mapper.check_last_error_code()

        py_reservation_list = []
        if ctypes_reservation_list:
            py_reservation_list = Reservation.to_py_reservation_list(ctypes_reservation_list)
            drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_reservation_list)))
        return py_reservation_list

    def get_all_jobs(self, filter):
        """ 
        Get jobs matching the specified info.

        :param filter: Job info filter.
        :type filter: JobInfo

        :returns: List of Job objects.

        >>> j_info = JobInfo({'job_name' : 'a_job'})
        >>> j_list = m_session.get_all_jobs(j_info)
        """
        self.logger.debug('Requesting list of jobs using filter: {}'.format(filter))
        drmaa2_lib = self.get_drmaa2_library()
        job_info = filter
        if type(filter) == PY_DICT_TYPE:
            job_info = JobInfo(filter)
        ctypes_filter = job_info._struct
        ctypes_job_list = drmaa2_lib.drmaa2_msession_get_all_jobs(self._struct, ctypes_filter)
        if not ctypes_job_list:
            self.exception_mapper.check_last_error_code()

        py_job_list = []
        if ctypes_job_list:
            py_job_list = Job.to_py_job_list(ctypes_job_list)
            drmaa2_lib.drmaa2_list_free(pointer(c_void_p(ctypes_job_list)))
        return py_job_list



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from ctypes import cast
from ctypes import POINTER

from .byte_string import ByteString
from .drmaa2_ctypes import drmaa2_slotinfo
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_object import Drmaa2Object


class SlotInfo(Drmaa2Object):
    """ High-level DRMAA2 slot info class. """

    machine_name = Drmaa2Object.StringDescriptor('machineName')
    """ Machine name (str). """
    slots = Drmaa2Object.LongLongDescriptor('slots')
    """ Number of slots (long). """
    implementation_specific = Drmaa2Object.ImplSpecDescriptor('implementationSpecific')
    """ Implementation specific dictionary ({str:str}). """

    def __init__(self, struct=None):
        """ Constructor. """
        Drmaa2Object.__init__(self)
        if struct is not None:
            self._struct = POINTER(drmaa2_slotinfo)()
            self._struct.contents = drmaa2_slotinfo()
            machine_name = ByteString(getattr(struct.contents, 'machineName').value).decode()
            # self.machine_name = ByteString(getattr(struct.contents, 'machineName').value).decode()
            self.machine_name = machine_name
            self.slots = getattr(struct.contents, 'slots')

    def __del__(self):
        pass

    @classmethod
    def get_implementation_specific_keys(cls):
        """
        Retrieve list of implementation-specific keys.

        :returns: String list of implementation-specific keys.
        """
        if cls.implementation_specific_keys is None:
            # cls.implementation_specific_keys = cls.to_py_string_list(cls.get_drmaa2_library().drmaa2_slotinfo_impl_spec())
            cls.implementation_specific_keys = []
        return cls.implementation_specific_keys

    @classmethod
    def to_py_job_list(cls, ctypes_list):
        py_job_list = list()
        if ctypes_list:
            count = cls.drmaa2_lib.drmaa2_list_size(ctypes_list)
            cls.logger.debug('Converting ctypes job list of size {}'.format(count))
            for i in range(count):
                void_ptr = cls.drmaa2_lib.drmaa2_list_get(ctypes_list, i)
                if void_ptr:
                    si = cast(void_ptr, POINTER(drmaa2_slotinfo))
                    si = SlotInfo(si)
                    py_job_list.append(si)
                else:
                    ExceptionMapper.check_last_error_code()
                    py_job_list.append(None)
        return py_job_list

    @classmethod
    def to_ctypes_job_list(cls, py_job_list):
        cls.logger.debug('Converting py job list of size {}'.format(len(py_job_list)))
        ctypes_job_list = cls.drmaa2_lib.drmaa2_list_create(int(ListType.SLOTINFOLIST), drmaa2_list_entryfree())
        for si in py_job_list:
            ExceptionMapper.check_status_code(cls.drmaa2_lib.drmaa2_list_add(ctypes_job_list, si._struct))
        return ctypes_job_list



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from ctypes import POINTER
from ctypes import cast
from ctypes import c_void_p
from ctypes import pointer

from .byte_string import ByteString
from .drmaa2_ctypes import drmaa2_version
from .drmaa2_ctypes import drmaa2_dict
from .drmaa2_object import Drmaa2Object


class Version(Drmaa2Object):
    """ High-level DRMAA2 version class. """

    major = Drmaa2Object.StringDescriptor('major')
    """ Major version string (str). """
    minor = Drmaa2Object.StringDescriptor('minor')
    """ Minor version string (str). """
    implementation_specific = Drmaa2Object.ImplSpecDescriptor('implementationSpecific')
    """ Implementation specific dictionary ({str:str}). """

    def __init__(self, version):
        """ 
        Constructor. 

        :param version: Low-level drmaa2_version struct.
        :type version: drmaa2_version
        """
        Drmaa2Object.__init__(self)
        if isinstance(version, POINTER(drmaa2_version)):
            self._struct = POINTER(drmaa2_version)()
            self._struct.contents = drmaa2_version()
            if version:
                self.major = ByteString(getattr(version.contents, 'major').value).decode()
                self.minor = ByteString(getattr(version.contents, 'minor').value).decode()
                self.implementation_specific = self.get_implementation_specific_attrs()
            else:
                self.major = "-NA-"
                self.minor = "-NA-"
        elif isinstance(version, drmaa2_version):
            self._struct = POINTER(drmaa2_version)()
            self._struct.contents = version
        else:
            raise InvalidArgument('Invalid argument: %s' % str(version))
        self._read_only = True

    def __del__(self):
        pass

    @classmethod
    def get_implementation_specific_attrs(cls):
        """
        Retrieve dict of implementation-specific attrs.

        :returns: {str:str} dict of implementation-specific attrs.

        >>> print(Version.get_implementation_specific_attrs())
        ['uge_version_json':'xyz']
        """
        ctypes_version = cls.get_drmaa2_library().drmaa2_get_drmaa_version();
        if cls.implementation_specific_attrs is None:
            cls.implementation_specific_attrs = cls.to_py_dict(
                cls.get_drmaa2_library().uge_vi_impl_spec_get(ctypes_version))
        return cls.implementation_specific_attrs

    @classmethod
    def get_implementation_specific_keys(cls):
        """
        Retrieve list of implementation-specific keys.

        :returns: String list of implementation-specific keys.

        >>> print(Version.get_implementation_specific_keys())
        ['uge_version_json']
        """
        if cls.implementation_specific_keys is None:
            cls.implementation_specific_keys = cls.to_py_string_list(
                cls.get_drmaa2_library().drmaa2_version_impl_spec())
        return cls.implementation_specific_keys

    @classmethod
    def get_drms_version(cls):
        """
        Retrieve DRMS version.

        :returns: DRMS version object.

        >>> v = Version.get_drms_version()
        >>> print(v.major)
        8
        >>> print(v.minor)
        6.0
        """
        ctypes_version = cls.get_drmaa2_library().drmaa2_get_drms_version();
        py_version = Version(ctypes_version)
        cls.get_drmaa2_library().drmaa2_version_free(ctypes_version)
        return py_version

    @classmethod
    def get_drmaa_version(cls):
        """
        Retrieve DRMAA version.

        :returns: DRMAA version object.

        >>> v = Version.get_drmaa_version()
        >>> print(v.major)
        2
        >>> print(v.minor)
        0
        """
        ctypes_version = cls.get_drmaa2_library().drmaa2_get_drmaa_version();
        py_version = Version(ctypes_version)
        cls.get_drmaa2_library().drmaa2_version_free(ctypes_version)
        return py_version



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from ctypes import cast
from ctypes import pointer
from ctypes import POINTER

from .drmaa2_constants import Bool
from .drmaa2_constants import Time
from .drmaa2_constants import JobState
from .drmaa2_constants import ListType
from .drmaa2_ctypes import drmaa2_string
from .drmaa2_ctypes import drmaa2_j
from .drmaa2_ctypes import drmaa2_list_entryfree

from .sudo import Sudo

from .byte_string import ByteString
from .job_info import JobInfo
from .job_template import JobTemplate
from .drmaa2_object import Drmaa2Object
from .log_manager import LogManager
from .exception_mapper import ExceptionMapper
from .drmaa2_exceptions import InvalidArgument


class Job(Drmaa2Object):
    """ High-level DRMAA2 job class. """

    id = Drmaa2Object.StringDescriptor('id')
    """ Job id (str). """
    session_name = Drmaa2Object.StringDescriptor('session_name')
    """ Session name (str). """
    job_name = Drmaa2Object.StringDescriptor('job_name')
    """ Job name (str). """

    logger = LogManager.get_instance().get_logger('Job')

    def __init__(self, job):
        """ Constructor. """
        Drmaa2Object.__init__(self)
        if isinstance(job, POINTER(drmaa2_j)):
            self._struct = POINTER(drmaa2_j)()
            self._struct.contents = drmaa2_j()
            self.id = ByteString(getattr(job.contents, 'id').value).decode()
            self.session_name = ByteString(getattr(job.contents, 'session_name').value).decode()
            self.job_name = ByteString(getattr(job.contents, 'job_name').value).decode()
        else:
            raise InvalidArgument('Invalid argument: %s' % str(job))

    def suspend(self, auth=None):
        """ 
        Suspend the job.

        :param auth: Optional sudo object for suspending the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.suspend()
        >>> print(j.get_info().job_state)
        SUSPENDED
        """
        self.logger.debug('Suspending job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_suspend_as(auth._struct, self._struct))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_suspend(self._struct))

    def resume(self, auth=None):
        """ 
        Resume the job.

        :param auth: Optional sudo object for resuming the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.suspend()
        >>> ...
        >>> j.resume()
        """
        self.logger.debug('Resuming job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_resume_as(auth._struct, self._struct))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_resume(self._struct))

    def hold(self, auth=None):
        """ 
        Hold the job.

        :param auth: Optional sudo object for holding the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.hold()
        >>> print(j.get_info().job_state)
        RUNNING_HELD
        """
        self.logger.debug('Holding job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_hold_as(auth._struct, self._struct))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_hold(self._struct))

    def release(self, auth=None):
        """ 
        Release the job.

        :param auth: Optional sudo object for releasing the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.hold()
        >>> ...
        >>> j.release()
        """
        self.logger.debug('Releasing job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_release_as(auth._struct, self._struct))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_release(self._struct))

    def terminate(self, auth=None):
        """ 
        Terminate the job.

        :param auth: Optional sudo object for terminating the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> j.terminate()
        >>> j.get_info().terminating_signal
        'SIGKILL'
        """
        self.logger.debug('Terminating job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            forced = Bool.FALSE
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_as(auth._struct, self._struct, forced))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate(self._struct))

    def terminate_forced(self, auth=None):
        """ 
        Terminate the job (forced).

        :param auth: Optional sudo object for terminating the job; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> j.terminate_forced()
        """
        self.logger.debug('Terminating (forced) job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            forced = Bool.TRUE
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_as(auth._struct, self._struct, forced))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_forced(self._struct))

    def terminate_all(self, auth=None):
        """ 
        Terminate all job tasks.

        :param auth: Optional sudo object for terminating tasks; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.terminate_all()
        """
        self.logger.debug('Terminating all tasks for job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            forced = Bool.FALSE
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_all_as(auth._struct, self._struct, forced))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_all(self._struct))

    def terminate_forced_all(self, auth=None):
        """ 
        Terminate all job tasks (forced).

        :param auth: Optional sudo object for terminating tasks; it can be specified either as a dictionary, or as a Sudo object directly.
        :type auth: Sudo or dict

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> j.terminate_forced_all()
        """
        self.logger.debug('Terminating all tasks (forced) for job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        if auth:
            auth = Sudo.create_from_dict(auth)
            self.logger.debug('Using sudo object: {}'.format(auth))
            forced = Bool.TRUE
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_all_as(auth._struct, self._struct, forced))
        else:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_terminate_forced_all(self._struct))

    def reap(self):
        """ 
        Reap the job from internal lists.

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> j.reap()
        """
        self.logger.debug('Reaping job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_reap(self._struct))

    def get_template(self):
        """ 
        Get job template.

        :returns: JobTemplate object.

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> jt = j.get_template()
        >>> print(jt.remote_command)
        /bin/sleep
        >>> print(jt.args)
        ['100']
        """
        self.logger.debug('Retrieving template for job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_template = drmaa2_lib.drmaa2_j_get_jtemplate(self._struct)
        if not ctypes_job_template:
            ExceptionMapper.check_last_error_code()

        return JobTemplate(ctypes_job_template)

    def get_state(self):
        """ 
        Get job state.

        :returns: (JobState object, sub-state string) tuple.

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> state,substate = j.get_state()
        >>> print(state)
        JobState.DONE
        """
        self.logger.debug('Retrieving state for job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        sub_state = POINTER(drmaa2_string)()
        sub_state.contents = drmaa2_string()
        jstate = drmaa2_lib.drmaa2_j_get_state(self._struct, sub_state)
        if jstate == int(JobState.UNSET_JSTATE):
            ExceptionMapper.check_last_error_code()
        return (JobState(jstate), sub_state.contents.value)

    def get_info(self):
        """ 
        Get job info.

        :returns: JobInfo object.

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> ...
        >>> ji = j.get_info()
        >>> print(ji.job_id)
        724.1
        >>> print(ji.slots)
        1
        """
        self.logger.debug('Retrieving info for job id {}'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        ctypes_job_info = drmaa2_lib.drmaa2_j_get_info(self._struct)
        if not ctypes_job_info:
            ExceptionMapper.check_last_error_code()
        return JobInfo(ctypes_job_info)

    def wait_started(self, timeout=Time.INFINITE_TIME):
        """ 
        Wait until the job starts or specified timeout occurs. 

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.wait_started()
        >>> print(j.get_state())
        (<JobState.RUNNING: 3>, None)
        """
        drmaa2_lib = self.get_drmaa2_library()
        self.logger.debug('Waiting on job id {} start'.format(self.id))
        ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_wait_started(self._struct, timeout))

    def wait_terminated(self, timeout=Time.INFINITE_TIME):
        """ 
        Wait until the job terminates or specified timeout occurs.

        :param timeout: Wait timeout in seconds (default: infinite time).
        :type timeout: int

        >>> j = j_session.run_job({'remote_command' : '/bin/sleep', 'args' : ['100']})
        >>> j.wait_terminated()
        >>> print(j.get_state())
        (<JobState.DONE: 3>, None)
        """
        self.logger.debug('Waiting on job id {} termination'.format(self.id))
        drmaa2_lib = self.get_drmaa2_library()
        ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_j_wait_terminated(self._struct, int(timeout)))

    def __del__(self):
        pass

    @classmethod
    def to_py_job_list(cls, ctypes_list):
        drmaa2_lib = cls.get_drmaa2_library()
        py_job_list = list()
        if ctypes_list:
            count = drmaa2_lib.drmaa2_list_size(ctypes_list)
            cls.logger.debug('Converting ctypes job list of size {}'.format(count))
            for i in range(count):
                void_ptr = drmaa2_lib.drmaa2_list_get(ctypes_list, i)
                if void_ptr:
                    j = cast(void_ptr, POINTER(drmaa2_j))
                    j = Job(j)
                    py_job_list.append(j)
                else:
                    ExceptionMapper.check_last_error_code()
                    py_job_list.append(None)
        return py_job_list

    @classmethod
    def to_ctypes_job_list(cls, py_job_list):
        drmaa2_lib = cls.get_drmaa2_library()
        cls.logger.debug('Converting py job list of size {}'.format(len(py_job_list)))
        ctypes_job_list = drmaa2_lib.drmaa2_list_create(int(ListType.JOBLIST), drmaa2_list_entryfree())
        for j in py_job_list:
            ExceptionMapper.check_status_code(drmaa2_lib.drmaa2_list_add(ctypes_job_list, j._struct))
        return ctypes_job_list



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

"""
Sudo class.
"""

from ctypes import POINTER
from .drmaa2_ctypes import drmaa2_sudo
from .drmaa2_object import Drmaa2Object


class Sudo(Drmaa2Object):
    """ High-level DRMAA2 sudo class. """

    username = Drmaa2Object.CharBufferDescriptor('username')
    """ Username to run operation as (str). """
    groupname = Drmaa2Object.CharBufferDescriptor('groupname')
    """ Groupname to run operation as (str). """
    uid = Drmaa2Object.LongDescriptor('uid')
    """ User id to run operation as (int). """
    gid = Drmaa2Object.LongDescriptor('gid')
    """ Group id to run operation as (int). """

    def __init__(self, username=None, groupname=None, uid=None, gid=None, sudo_dict={}):
        """ 
        Constructor. 

        :param username: Sudo object username; it may be also be specified as part of the sudo object dictionary.
        :type username: str
 
        :param groupname: Sudo object groupname; it may be also be specified as part of the sudo object dictionary.
        :type groupname: str
 
        :param uid: Sudo object user id; it may be also be specified as part of the sudo object dictionary.
        :type uid: int
 
        :param gid: Sudo object group id; it may be also be specified as part of the sudo object dictionary.
        :type gid: int
 
        :param sudo_dict: Sudo object dictionary; if specified, it should contain 'username', 'groupname', 'uid' and/or 'gid' keys.
        :type sudo_dict: dict

        >>> auth = Sudo(username='auser', groupname='agroup')
        >>> print('Username: %s' % auth.username)
        >>> Username: auser
        """
        Drmaa2Object.__init__(self)
        self._struct = POINTER(drmaa2_sudo)()
        self._struct.contents = drmaa2_sudo()

        username = username or sudo_dict.get('username')
        self.username = username
        self.groupname = groupname or sudo_dict.get('groupname')
        self.uid = uid or sudo_dict.get('uid')
        self.gid = gid or sudo_dict.get('gid')

    def __del__(self):
        pass



# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__
from .drmaa2_constants import (Bool, Capability, Cpu, Event, JobState, 
    ListType, Os, ResourceLimit, StatusCode, Time)
from .drmaa2_exceptions import (Drmaa2Exception, DeniedByDrms,
    DrmCommunicationError, TryLaterError, SessionManagementError, 
    TimeoutError, InternalError, InvalidArgument, 
    InvalidSession, InvalidState, ResourceNotAvailable, 
    UnsupportedAttribute, UnsupportedOperation,
    ImplementationSpecificError, AuthorizationError)
from .library_manager import LibraryManager
from .log_manager import LogManager
from .job_session import JobSession
from .job_info import JobInfo
from .job_array import JobArray
from .job_template import JobTemplate
from .job import Job
from .reservation_session import ReservationSession
from .reservation_template import ReservationTemplate
from .reservation_info import ReservationInfo
from .reservation import Reservation
from .sudo import Sudo
from .version import Version
from .monitoring_session import MonitoringSession
from .machine_info import MachineInfo
from .queue_info import QueueInfo
from .notification import Notification

get_drms_name = LibraryManager.get_drms_name
get_drmaa_name = LibraryManager.get_drmaa_name
drmaa_supports = LibraryManager.drmaa_supports 
get_drms_version = Version.get_drms_version
get_drmaa_version = Version.get_drmaa_version
get_job_session_names = JobSession.list_session_names 
get_reservation_session_names = ReservationSession.list_session_names 
 
__version__ = '8.7.0'



#!/usr/bin/env python
# ___INFO__MARK_BEGIN__
#######################################################################################
# Copyright 2008-2021 Univa Corporation (acquired and owned by Altair Engineering Inc.)
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#######################################################################################
# ___INFO__MARK_END__

from ctypes import POINTER
from ctypes import cast

from .byte_string import ByteString
from .drmaa2_constants import Cpu
from .drmaa2_constants import Os
from .drmaa2_ctypes import drmaa2_queueinfo
from .drmaa2_object import Drmaa2Object
from .drmaa2_exceptions import InvalidArgument

from .drmaa2_version_descriptor import Drmaa2VersionDescriptor


class QueueInfo(Drmaa2Object):
    """ High-level DRMAA2 queue info class. """

    name = Drmaa2Object.StringDescriptor('name')
    """ Queue name (str). """
    implementation_specific = Drmaa2Object.ImplSpecDescriptor('implementationSpecific')
    """ Implementation specific dictionary ({str:str}). """

    def __init__(self, queue_info):
        """ 
        Constructor. 

        :param queue_info: Low-level drmaa2_queueinfo struct.
        :type queue_info: drmaa2_queueinfo
        """
        Drmaa2Object.__init__(self)
        if isinstance(queue_info, POINTER(drmaa2_queueinfo)):
            self._struct = POINTER(drmaa2_queueinfo)()
            self._struct.contents = drmaa2_queueinfo()
            self.name = ByteString(getattr(queue_info.contents, 'name').value).decode()
            self.implementation_specific = getattr(queue_info.contents, 'implementationSpecific')
        else:
            raise InvalidArgument('Invalid argument: %s' % str(queue_info))
        self._read_only = True

    def __del__(self):
        pass

    @classmethod
    def get_implementation_specific_keys(cls):
        """
        Retrieve list of implementation-specific keys.

        :returns: String list of implementation-specific keys.

        >>> print(QueueInfo.get_implementation_specific_keys())
        []
        """
        if cls.implementation_specific_keys is None:
            cls.implementation_specific_keys = cls.to_py_string_list(
                cls.get_drmaa2_library().drmaa2_queueinfo_impl_spec())
        return cls.implementation_specific_keys

    @classmethod
    def to_py_queue_info_list(cls, ctypes_list):
        drmaa2_lib = cls.get_drmaa2_library()
        py_queue_info_list = list()
        if ctypes_list:
            count = drmaa2_lib.drmaa2_list_size(ctypes_list)
            cls.logger.debug('Converting ctypes queue info list of size {}'.format(count))
            for i in range(count):
                void_ptr = drmaa2_lib.drmaa2_list_get(ctypes_list, i)
                if void_ptr:
                    qi = cast(void_ptr, POINTER(drmaa2_queueinfo))
                    qi = QueueInfo(qi)
                    py_queue_info_list.append(qi)
                else:
                    ExceptionMapper.check_last_error_code()
                    py_queue_info_list.append(None)
        return py_queue_info_list
